#ifndef DUNE_MICROSTRUCTURE_EFFECTIVEQUANTITIESCOMPUTER_HH
#define DUNE_MICROSTRUCTURE_EFFECTIVEQUANTITIESCOMPUTER_HH

#include <filesystem>


#include <dune/microstructure/matrix_operations.hh>
#include <dune/microstructure/CorrectorComputer.hh>

using namespace MatrixOperations;
using std::shared_ptr;
using std::make_shared;
using std::string;
using std::cout;
using std::endl;

// template <class Basis>
// class EffectiveQuantitiesComputer : public CorrectorComputer<Basis> {

template <class Basis>
class EffectiveQuantitiesComputer {

public:
	static const int dimworld = 3;
	// static const int nCells = 4;
    

	static const int dim = Basis::GridView::dimension;

	using Domain = typename CorrectorComputer<Basis>::Domain; 

	using VectorRT = typename CorrectorComputer<Basis>::VectorRT;
	using MatrixRT = typename CorrectorComputer<Basis>::MatrixRT;

	using Func2Tensor = typename CorrectorComputer<Basis>::Func2Tensor;
	using FuncVector = typename CorrectorComputer<Basis>::FuncVector;

	using VectorCT = typename CorrectorComputer<Basis>::VectorCT;

	using HierarchicVectorView = typename CorrectorComputer<Basis>::HierarchicVectorView;

protected:

	CorrectorComputer<Basis>& correctorComputer_; 
	Func2Tensor prestrain_;

public:
	VectorCT B_load_TorusCV_;				//<B, Chi>_L2 
	FieldMatrix<double, dim, dim> Q_;	    //effective moduli <LF_i, F_j>_L2
	FieldVector<double, dim> Bhat_;			//effective loads induced by prestrain <LF_i, B>_L2
	FieldVector<double, dim> Beff_;		//effective strains Mb = ak

	// corrector parts
	VectorCT phi_E_TorusCV_;		//phi_i * (a,K)_i
	VectorCT phi_perp_TorusCV_;
	VectorCT phi_TorusCV_;
	VectorCT phi_1_;		//phi_i * (a,K)_i
	VectorCT phi_2_;
	VectorCT phi_3_;
	
	// is this really interesting???
	// double phi_E_L2norm_;
	// double phi_E_H1seminorm_;

	// double phi_perp_L2norm_;
	// double phi_perp_H1seminorm_;

	// double phi_L2norm_;
	// double phi_H1seminorm_;

	// double Chi_E_L2norm_;		
	// double Chi_perp_L2norm_;
	// double Chi_L2norm_;


	double B_energy_;			 // < B, B >_L 		B = F + Chi_perp + B_perp 
	double F_energy_;			 // < F, F >_L
	double Chi_perp_energy_;	 // < Chi_perp, Chi_perp >_L
	double B_perp_energy_; 		 // < B_perp, B_perp >_L

	//Chi(phi) is only implicit computed, can we store this?

 
	///////////////////////////////
	// constructor
	///////////////////////////////
	// EffectiveQuantitiesComputer(CorrectorComputer<Basis>& correctorComputer, Func2Tensor prestrain)
    //     : correctorComputer_(correctorComputer), prestrain_(prestrain)
	EffectiveQuantitiesComputer(CorrectorComputer<Basis>& correctorComputer)
        : correctorComputer_(correctorComputer)
    { 
    	
    	// computePrestressLoadCV();
	  	// computeEffectiveStrains();

    	// compute_phi_E_TorusCV();
    	// compute_phi_perp_TorusCV();
    	// compute_phi_TorusCV();

    	// computeCorrectorNorms();
    	// computeChiNorms();
    	// computeEnergiesPrestainParts();	

    	// writeInLogfile();
    } 


private:

    void computeQ()
    {

        auto test = correctorComputer_.getLoad_alpha1();
        auto phiContainer = correctorComputer_.getPhicontainer();

        // std::cout << mu_ << "mu_ can be used this way"
        // auto mu = 
        // auto lamba = 
        // auto gamma = 

        return ;
    }









	// void computePrestressLoadCV()
    // {
    // 	cout << "computePrestressLoadCV ..." << endl;
    // 	auto cellAssembler = cellSolver_.getCellAssembler();
	//   	cellAssembler.assembleLoadVector(prestrain_, B_load_TorusCV_);
    // }

    // void computeEffectiveStrains()
    // {
    // 	cout << "Compute effective strains ..." << endl;
    // 	auto cellAssembler = cellSolver_.getCellAssembler();
	
	//   	//E_h = (U +  K e_cs, 0, 0)
		
	//   	// lateral stretch strain		E_U1 = e1 ot e1  				= MatrixRT{{1, 0, 0}, {0, 0, 0}, {0, 0, 0}};
	// 	Func2Tensor E_a  = [&] (const Domain& z) { return biotStrainApprox({1, 0, 0}, {0, 0, 0}, {0, z[dim-2], z[dim-1]}); }; 

	//   	// twist in x2-x3 plane 		E_K1 = (e1 x (0,x2,x3)^T ot e1) = MatrixRT{{0,-z[2], z[1]}, {0, 0 , 0 }, {0,0,0}}; 
	//     Func2Tensor E_K1 = [&] (const Domain& z) { return biotStrainApprox({0, 0, 0}, {1, 0, 0}, {0, z[dim-2], z[dim-1]}); };

	//     //  bend strain in x1-x2 plane	E_K2 = (e2 x (0,x2,x3)^T ot e1) = MatrixRT{{z[2], 0, 0}, {0, 0, 0}, {0, 0, 0}}; 
	//   	Func2Tensor E_K2 = [&] (const Domain& z) { return biotStrainApprox({0, 0, 0}, {0, 1, 0}, {0, z[dim-2], z[dim-1]}); };
	  	
	//   	// bend strain in x1-x3 plane	E_K3 = (e3 x (0,x2,x3)^T ot e1) = MatrixRT{{-z[1], 0, 0}, {0, 0, 0}, {0, 0, 0}};
	//   	Func2Tensor E_K3 = [&] (const Domain& z) { return biotStrainApprox({0, 0, 0}, {0, 0, 1}, {0, z[dim-2], z[dim-1]});  };
	  	
	  	
	// 	shared_ptr<VectorCT> phi_Ei_TorusCV[nCells] = {cellSolver_.getCorr_a(), 
	// 												cellSolver_.getCorr_K1(), 
	// 												cellSolver_.getCorr_K1(),
	// 												cellSolver_.getCorr_K1()};

	//   	shared_ptr<VectorCT> loadsCell[nCells] = {cellAssembler.getLoad_a(),
	//   											  cellAssembler.getLoad_K1(), 
	//   											  cellAssembler.getLoad_K2(), 
	//   											  cellAssembler.getLoad_K3()};

	//   	const std::array<shared_ptr<Func2Tensor>, nCells> strainsE = {make_shared<Func2Tensor>(E_a), 
	// 																  make_shared<Func2Tensor>(E_K1), 
	// 																  make_shared<Func2Tensor>(E_K2),
	// 																  make_shared<Func2Tensor>(E_K3)};	
	  	
	  
  	// 	for (unsigned int k = 0; k < nCells; k++)
  	// 		for (unsigned int l = 0; l < nCells; l++){
  	// 			std::cout << "<L Fi,Fj> ... " << k << l << std::endl;
  	// 			M_[k][l] = cellAssembler.energyScalarProduct_FiFj(*strainsE[k], *phi_Ei_TorusCV[k], *strainsE[l], *phi_Ei_TorusCV[l]);
  	// 		}


	//   	for (unsigned int k = 0; k < nCells; k++){
	//   		std::cout << "<L Fi,B> ... " << k << std::endl;
	//   		b_[k] = cellAssembler.energyScalarProduct_FiB(*strainsE[k], *phi_Ei_TorusCV[k], prestrain_);
	//   	}
	  		

	//   	M_.solve( aK_ ,b_);
	//   	std::cout << "\nstretch twist bend1 bend2\naK = "<< aK_ << std::endl;
    // }

    // void compute_phi_E_TorusCV()
	// {
	// 	cout << "compute_phi_E_TorusCV ..." << endl;
	// 	phi_E_TorusCV_ = *(cellSolver_.getCorr_a());
	//     phi_E_TorusCV_ *= aK_[0];
	//     auto tmp_xK1 = *(cellSolver_.getCorr_K1());
	//     auto tmp_xK2 = *(cellSolver_.getCorr_K2());
	//     auto tmp_xK3 = *(cellSolver_.getCorr_K3());
	//     tmp_xK1 *= aK_[1];
	//     tmp_xK2 *= aK_[2];
	//     tmp_xK3 *= aK_[3];
	//     phi_E_TorusCV_ += tmp_xK1;
	//     phi_E_TorusCV_ += tmp_xK2;
	//     phi_E_TorusCV_ += tmp_xK3;
	// }

    // void compute_phi_perp_TorusCV()
    // {
    // 	cout << "compute_phi_perp_TorusCV ..." << endl;
	//   	cellSolver_.solvePrestrainCorrector(phi_perp_TorusCV_, B_load_TorusCV_);
    // }

    // void compute_phi_TorusCV()
    // {
    // 	cout << "compute_phi_TorusCV ..." << endl;
	//   	phi_TorusCV_.resize(phi_E_TorusCV_.size());
	//   	for (int i=0; i < phi_TorusCV_.size(); i++)
	//   		phi_TorusCV_[i] = phi_E_TorusCV_[i] + phi_perp_TorusCV_[i];
    // }







	
	public:
	///////////////////////////////
	// getter
	///////////////////////////////
	CorrectorComputer<Basis> getCorrectorComputer(){return correctorComputer_;}

	const shared_ptr<Basis> getBasis()  
	{
		return correctorComputer_.getBasis();
	}




	


	  	

}; // end class





#endif